/*
 * Copyright (C) 1997-2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "server.h"
#include "config.h"
#include <sys/mman.h> /* MAP_SHARED, PROT_READ, PROT_WRITE */

#define BACKLOG 10 /*  how many pending connections queue will hold */


/* these are exported to afs/command/dbtool */
struct misc_meta_data *mmd;
struct gengetopt_args_info conf;
char *user_list;

/* list of database tools */
struct dbtool dblist[] = {
	{
		.name = "dopey",
		.init = dopey_dbtool_init,
	},
#if HAVE_MYSQL
	{
		.name = "mysql",
		.init = mysql_dbtool_init,
	},
#endif
	{
		.name = NULL,
	}
};


/* global variables for server-internal use */
static FILE *logfile;
static int mmd_semid;
static int signal_pipe[2];

/*
 * setup shared memory area and get semaphore for locking
 */
static void shm_init(void)
{
	int fd;
	caddr_t area;

	if ((fd = open("/dev/zero", O_RDWR)) < 0) {
		clog(EMERG, "%s: failed to open /dev/zero\n", __func__);
		exit(EXIT_FAILURE);
	}
	if ((area = mmap(0, sizeof(struct misc_meta_data),
			PROT_READ | PROT_WRITE,
			MAP_SHARED, fd, 0)) == (caddr_t) - 1) {
		clog(EMERG, "%s: mmap error\n", __func__);
		exit(EXIT_FAILURE);
	}
	close(fd); /* we dont need /dev/zero anymore */
	mmd = (struct misc_meta_data *)area;

	mmd->dbt_num = 0;
	mmd->num_played = 0;
	mmd->num_commands = 0;
	mmd->events = 0;
	mmd->num_connects = 0;
	mmd->active_connections = 0;
	strcpy(mmd->filename, "(none)");
	mmd->audio_format = -1;
	mmd->afs_status_flags = AFS_NEXT;
	mmd->new_afs_status_flags = AFS_NEXT;

	mmd_semid = semget(42, 1, IPC_CREAT | 0666);
	if (mmd_semid == -1) {
		clog(EMERG, "%s: semget failed\n", __func__);
		exit(EXIT_FAILURE);
	}
}

static void do_semop(struct sembuf *sops, int num)
{
	if (semop(mmd_semid, sops, num) >= 0)
		return;
	clog(WARNING, "%s: semop failed (%s), retrying\n", __func__,
		strerror(errno));
	while (semop(mmd_semid, sops, num) < 0)
		; /* nothing */
}

/*
 * lock shared memory area
 */
void mmd_lock(void)
{
	struct sembuf sops[2] = {
		{
			.sem_num = 0,
			.sem_op = 0,
			.sem_flg = SEM_UNDO
		},
		{
			.sem_num = 0,
			.sem_op = 1,
			.sem_flg = SEM_UNDO
		}
	};
	do_semop(sops, 2);
}

/*
 * unlock shared memory area
 */
void mmd_unlock(void)
{
	struct sembuf sops[1] = {
		{
			.sem_num = 0,
			.sem_op = -1,
			.sem_flg = SEM_UNDO
		},
	};
	do_semop(sops, 1);
}

/*
 * set or get server's uptime
 */
int server_uptime(int set_or_get)
{
	static time_t startuptime;
	time_t now;

	if (set_or_get == SET) {
		time(&startuptime);
		return 0;
	}
	time(&now);
	return((int) difftime(now, startuptime));
}

void clog(int, char*, ...) __attribute__((format (printf, 2, 3)));
/*
 * log function. first argument is loglevel.
 */
void clog(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";
	pid_t mypid;
	char *dummy;

	if (ll < conf.loglevel_arg)
		return;
	if (!logfile) {
		if (ll < WARNING)
			outfd = stdout;
		else
			outfd = stderr;
	} else
		outfd = logfile;
	if (conf.daemon_given && !logfile)
		return;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s ", str);
	if (conf.loglevel_arg <= VERBOSE)
		fprintf(outfd, "%i: ", ll);
	mypid = getpid();
	if (conf.loglevel_arg <= VERBOSE)
		fprintf(outfd, "(%d) ", mypid);
	if (conf.loglevel_arg > VERBOSE) {
		fmt = strchr(fmt, ':');
		if (!fmt)
			return;
		fmt++;
		fmt++;
	}
	va_start(argp, fmt);
	if (conf.loglevel_arg > VERBOSE)
		dummy = va_arg(argp, char*);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

static void parse_config(int override)
{
	char *home = para_homedir();
	struct gengetopt_args_info *args_info = &conf;
	static char default_config_file[_POSIX_PATH_MAX];
	static char default_user_list[_POSIX_PATH_MAX];
	struct stat statbuf;
	int ret;
	char *cf = default_config_file;

	if (args_info->config_file_given)
		cf = args_info->config_file_arg;
	else
		sprintf(cf, "%s/.paraslash/server.conf", home);
	ret = stat(cf, &statbuf);
	if (ret && args_info->config_file_given) {
		clog(EMERG, "%s can not stat config file %s\n", __func__, cf);
		exit(EXIT_FAILURE);
	}
	if (!ret) {
		int tmp = args_info->daemon_given;
		cmdline_parser_configfile(cf, args_info, override, 0, 0);
		args_info->daemon_given = tmp;
	}
	/* user list */
	if (!args_info->user_list_given) {
		user_list = default_user_list;
		sprintf(user_list, "%s/.paraslash/server.users", home);
	} else
		user_list = args_info->user_list_arg;
	/* logfile */
	if (!args_info->logfile_given && args_info->daemon_given) {
		clog(EMERG, "%s daemon, but no config file\n", __func__);
		exit(EXIT_FAILURE);
	}
	if (args_info->logfile_given)
		logfile = open_log(args_info->logfile_arg);
	free(home);
}

static void setup_signal_handling(void)
{
	int ret = 0;

	para_signal_init(signal_pipe, clog);
	fcntl(signal_pipe[0], F_SETFL, O_NONBLOCK);
	clog(NOTICE, "%s: setting up signal handlers\n", __func__);
	ret += para_install_sighandler(SIGINT);
	ret += para_install_sighandler(SIGTERM);
	ret += para_install_sighandler(SIGHUP);
	ret += para_install_sighandler(SIGCHLD);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGUSR1, SIG_IGN);
	if (ret != 4) {
		clog(EMERG, "%s: could not install signal handlers\n",
			__func__);
		exit(EXIT_FAILURE);
	}
}

void init_dbtool()
{
	int i;

	mmd->dbt_change = -1; /* no change nec., set to new dbt num by com_cdt */
	clog(NOTICE, "%s: initializing database tool\n", __func__);
	if (!dblist[1].name)
		goto dopey;
	if (conf.dbtool_given) {
		for (i = 0; dblist[i].name; i++) {
			if (strcmp(dblist[i].name, conf.dbtool_arg))
				continue;
			if (dblist[i].init(&dblist[i]) < 0) {
				clog(WARNING, "%s: init %s failed",
					dblist[i].name, __func__);
				goto dopey;
			}
			mmd->dbt_num = i;
			return;
		}
		clog(WARNING, "%s: no such dbtool, switching to dopey\n",
			__func__);
		goto dopey;
	}
	/* use the first dbtool that works
	 * (assuming that dopey always works)
	 */
	for (i = 1; dblist[i].name; i++) {
		if (dblist[i].init(&dblist[i]) >= 0) {
			clog(INFO, "%s: initialized %s\n", __func__,
				dblist[i].name);
			mmd->dbt_num = i;
			return;
		}
		clog(CRIT, "%s: %s init failed\n", __func__,
			dblist[i].name);
	}
dopey:
	mmd->dbt_num = 0;
	dblist[0].init(&dblist[0]); /* always successful */
}

static int init_network(void)
{
	int sockfd;
	struct sockaddr_in my_addr;

	clog(INFO, "%s: init network socket\n", __func__);
	if ((sockfd = get_socket()) < 0)
		exit(EXIT_FAILURE);
	if (setserversockopts(sockfd) < 0)
		exit(EXIT_FAILURE);
	init_sockaddr(&my_addr, conf.port_arg, NULL);
	if (do_bind(sockfd, &my_addr) < 0)
		exit(EXIT_FAILURE);
	if (listen(sockfd, BACKLOG) == -1) {
		clog(EMERG, "%s: listen failed\n", __func__);
		exit(EXIT_FAILURE);
	}
	return sockfd;
}

static int do_inits(int argc, char **argv)
{
	/* connector's address information */
	int sockfd;

	/* parse command line options */
	cmdline_parser(argc, argv, &conf);
	/* parse config file, open log and set defaults */
	parse_config(0);
	/* print/log welcome message to console */
	if (conf.loglevel_arg < NOTICE)
		cmdline_parser_print_version();
	log_welcome("para_server", conf.loglevel_arg);
	shm_init(); /* init mmd struct */
	server_uptime(SET); /* reset server uptime */
	/* become daemon */
	if (conf.daemon_given)
		daemon_init();
	init_dbtool();
	clog(NOTICE, "%s: initializing audio file sender\n", __func__);
	/* audio file sender */
	if (afs_init(&conf) < 0)
		exit(EXIT_FAILURE);
	mmd->server_pid = getpid();
	setup_signal_handling();
	mmd_lock();
	/* init network socket */
	clog(NOTICE, "%s: initializing networking\n", __func__);
	sockfd = init_network();
	if (conf.autoplay_given) {
		mmd->afs_status_flags |= AFS_PLAYING;
		mmd->new_afs_status_flags |= AFS_PLAYING;
	}
	clog(NOTICE, "%s: init complete\n", __func__);
	return sockfd;
}

static void handle_dbt_change(void)
{
	int ret, old = mmd->dbt_num, new = mmd->dbt_change;

	dblist[old].shutdown();
	ret = dblist[new].init(&dblist[new]);
	mmd->dbt_change = -1; /* reset */
	if (ret >= 0) {
		mmd->dbt_num = new;
		return;
	}
	/* switch to dopey since init failed */
	dblist[0].init(&dblist[0]);
	mmd->dbt_num = 0;
}

/*
 * called when server gets SIGHUP or when client invokes hup command.
 */
static int handle_sighup(void)
{
	clog(NOTICE, "%s: SIGHUP\n", __func__);
	close_log(logfile); /* gets reopened if necessary by parse_config */
	logfile = NULL;
	parse_config(1); /* reopens log */
	mmd->dbt_change = mmd->dbt_num; /* do not change dbtool */
	handle_dbt_change(); /* force reloading dbtool */
	if (afs_handle_sighup(&conf) < 0)
		exit(EXIT_FAILURE);
	return 0;
}

static int need_status_refresh(void)
{
	static int prev_uptime = -1, prev_events = -1;
	int uptime = server_uptime(GET), ret = 1;

	if (prev_events != mmd->events)
		goto out;
	if (mmd->new_afs_status_flags != mmd->afs_status_flags)
		goto out;
	if (uptime / 60 != prev_uptime / 60)
		goto out;
	ret = 0;
out:
	prev_uptime = uptime;
	prev_events = mmd->events;
	mmd->afs_status_flags = mmd->new_afs_status_flags;
	return ret;
}

/*
 * MAIN
 */
int main(int argc, char **argv)
{
	/* listen on sock_fd, new connection on new_fd */
	int sockfd, new_fd, prev_events = -1;
	struct sockaddr_in their_addr;
	int err, i, sw_fileno, max_fileno, ret, check_stream_writer;
	pid_t chld_pid;
	fd_set rfds, wfds;
	struct timeval *timeout = NULL;

	sockfd = do_inits(argc, argv);
repeat:
	/* check socket and signal pipe in any case */
	FD_ZERO(&rfds);
	FD_SET(sockfd, &rfds);
	max_fileno = sockfd;
	FD_SET(signal_pipe[0], &rfds);
	max_fileno = MAX(max_fileno, signal_pipe[0]);

	check_stream_writer = 0;
	/* do we need to send a chunk of audio data? */
	sw_fileno = afs_mainloop(&timeout);
	if (sw_fileno >= 0) {
		check_stream_writer = 1;
		FD_ZERO(&wfds);
		FD_SET(sw_fileno, &wfds);
		max_fileno = MAX(sw_fileno, max_fileno);
	}
	if (need_status_refresh())
		killpg(0, SIGUSR1);
	prev_events = mmd->events;

	mmd_unlock();
//	clog(DEBUG, "%s: select (max = %i)\n", __func__, max_fileno);
	ret = select(max_fileno + 1, &rfds, check_stream_writer? &wfds : NULL,
		NULL, timeout? timeout : NULL);
	err = errno;
	//clog(DEBUG, "%s: select returned  %i\n", __func__, ret);
	mmd_lock();
	if (mmd->dbt_change >= 0)
		handle_dbt_change();
	if (ret < 0 && err == EINTR)
		goto repeat;
	if (ret < 0) {
		clog(CRIT, "%s: select error (%s)\n", __func__, strerror(err));
		goto repeat;
	}
	if (check_stream_writer && FD_ISSET(sw_fileno, &wfds))
		afs_send_chunk();
	if (FD_ISSET(signal_pipe[0], &rfds)) {
		int sig;
		pid_t pid;
		sig = para_next_signal();
		switch (sig) {
		case SIGHUP:
			handle_sighup();
			break;
		case SIGCHLD:
			while ((pid = para_reap_child()) > 0)
				afs_handle_sigchild(pid);
			break;
		/* die on sigint/sigterm. Kill all children too. */
		case SIGINT:
		case SIGTERM:
			clog(EMERG, "%s: terminating on signal %d\n",
				__func__, sig);
			kill(0, SIGTERM);
			exit(EXIT_FAILURE);
		}
	}
	if (!FD_ISSET(sockfd, &rfds))
		goto repeat;

	new_fd = do_accept(sockfd, &their_addr);
	if (new_fd < 0)
		goto repeat;
	clog(NOTICE, "%s: got connection from %s, forking\n", __func__,
		inet_ntoa(their_addr.sin_addr));
	mmd->num_connects++;
	mmd->active_connections++;
	chld_pid = fork();
	if (chld_pid < 0) {
		clog(CRIT, "%s: fork failed\n", __func__);
		goto repeat;
	}
	if (chld_pid) {
		close(new_fd);
		/* parent keeps accepting connections */
		goto repeat;
	}
	for (i = 3; i < 256; i++) { /* FIXME */
		int flags;
		if (i == new_fd)
			continue;
		flags = fcntl(i, F_GETFD);
		if (flags > 0 && flags & FD_CLOEXEC) {
			clog(DEBUG, "%s: closing fd %d\n", __func__, i);
			close(i);
		}
	}
	close(sockfd); /* child doesn't need the listener */
	/*
	 * put info on who we are serving into argv[0] to make
	 * client ip visible in top/ps
	 */
	for (i = argc - 1; i >= 0; i--)
		memset(argv[i], 0, strlen(argv[i]));
	sprintf(argv[0], "para_server (serving %s)",
		inet_ntoa(their_addr.sin_addr));
	return handle_connect(new_fd);
}
